from pathlib import Path
from ansicolor import Colors, colorize


DEFAULT_FORMAT = '{file_path}:{line_number}:{column_number}: {description} (see {reference})'

FORMAT_COLOR_MAP = {
    'file_path': Colors.Red,
    'file_name': Colors.Red,
    'line_number': Colors.White,
    'column_number': Colors.White,
    'severity': Colors.Red,
    'description': Colors.White,
    'policy_name': Colors.White,
    'reference': Colors.White,
}


class Formatter(object):
    def __init__(self, config_dict):
        if 'cmdargs' in config_dict:
            cmdargs = config_dict['cmdargs']
        else:
            cmdargs = {}

        if 'format' in cmdargs and cmdargs['format'] is not None:
            self._format = cmdargs['format']
        else:
            self._format = DEFAULT_FORMAT

        if 'color' in cmdargs and cmdargs['color'] is not None:
            self._should_be_colorized = cmdargs['color']
        else:
            self._should_be_colorized = False


    def _sort_violations(self, violations):
        return sorted(violations, key=lambda violation: (violation['position']['path'],
                                                         violation['position']['line']))


    def format_violations(self, violations):
        sorted_violations = self._sort_violations(violations)
        formatted_lines = map(self.format_violation, sorted_violations)

        return '\n'.join(formatted_lines)


    def format_violation(self, violation):
        if self._should_be_colorized:
            formatter_map = self._get_colorize_formatter_map(violation)
        else:
            formatter_map = self._get_formatter_map(violation)

        formatted_line = self._format.format(**formatter_map)
        return formatted_line


    def _get_colorize_formatter_map(self, violation):
        formatter_map = self._get_formatter_map(violation)
        colorized_formatter_map = {}

        for key, value in formatter_map.items():
            if key in FORMAT_COLOR_MAP:
                Color = FORMAT_COLOR_MAP[key]
                colorized_formatter_map[key] = colorize(str(value), Color())
            else:
                colorized_formatter_map[key] = value

        return colorized_formatter_map


    def _get_formatter_map(self, violation):
        file_path = Path(violation['position']['path'])

        return {
            'file_path': file_path,
            'file_name': file_path.name,
            'line_number': violation['position']['line'],
            'column_number': violation['position']['column'],
            'severity': violation['level'].name.lower(),
            'description': violation['description'],
            'policy_name': violation['name'],
            'reference': violation['reference'],
        }



import logging
from vint.linting.level import is_level_enabled
import vint.linting.policy_registry as policy_registry


class PolicySet(object):
    def __init__(self):
        self._all_policies_map = self._create_all_policies_map()
        self.enabled_policies = []


    def _create_all_policies_map(self):
        policy_class_map = policy_registry.get_policy_class_map()

        policy_map = {policy_name: PolicyClass()
                      for policy_name, PolicyClass
                      in policy_class_map.items()}

        return policy_map


    def _is_policy_exists(self, name):
        return name in self._all_policies_map


    def _get_policy(self, name):
        return self._all_policies_map[name]


    def _warn_unexistent_policy(self, policy_name):
        logging.warning('Policy `{name}` is not defined'.format(
            name=policy_name))


    def _log_severity(self, config_dict):
        level_name = config_dict['cmdargs']['severity'].name
        logging.debug('severity: {level}'.format(level=level_name))


    def _get_enabling_map(self, config_dict):
        self._log_severity(config_dict)

        severity = config_dict['cmdargs']['severity']
        policy_enabling_map = {}

        for policy_name, policy in self._all_policies_map.items():
            policy_enabling_map[policy_name] = is_level_enabled(policy.level, severity)

        prior_policy_enabling_map = config_dict['policies']

        for policy_name, policy in prior_policy_enabling_map.items():
            if 'enabled' in policy:
                policy_enabling_map[policy_name] = policy['enabled']

        return policy_enabling_map


    def update_by_config(self, config_dict):
        """ Update policies set by the config dictionary.

        Expect the policy_enabling_map structure to be (represented by YAML):
            - PolicyFoo:
              enabled: True
            - PolicyBar:
              enabled: False
              additional_field: 'is_ok'
        """
        policy_enabling_map = self._get_enabling_map(config_dict)
        self.enabled_policies = []

        for policy_name, is_policy_enabled in policy_enabling_map.items():
            if not self._is_policy_exists(policy_name):
                self._warn_unexistent_policy(policy_name)
                continue

            if is_policy_enabled:
                enabled_policy = self._get_policy(policy_name)
                self.enabled_policies.append(enabled_policy)

            self._log_policy_status(policy_name, is_policy_enabled)


    def get_enabled_policies(self):
        """ Returns enabled policies. """
        return self.enabled_policies


    def _log_policy_status(self, policy_name, enabled):
        status = 'enabled' if enabled else 'disabled'
        logging.debug('{status}: {policy_name}'.format(status=status,
                                                       policy_name=policy_name))



import sys
from typing import Dict, Any
from pprint import pformat
from pathlib import Path
from vint.encodings.decoding_strategy import DecodingStrategy


SCRIPTENCODING_PREFIX = bytearray('scriptencoding', encoding='ascii')



class Decoder(object):
    def __init__(self, strategy):
        # type: (DecodingStrategy) -> None
        self.strategy = strategy
        self.debug_hint = dict(version=sys.version)


    def read(self, file_path):
        # type: (Path) -> str

        with file_path.open(mode='rb') as f:
            bytes_seq = f.read()
            strings = []

            for (loc, hunk) in _split_by_scriptencoding(bytes_seq):
                debug_hint_for_the_loc = dict()
                self.debug_hint[loc] = debug_hint_for_the_loc

                string = self.strategy.decode(hunk, debug_hint=debug_hint_for_the_loc)

                if string is None:
                    raise EncodingDetectionError(self.debug_hint)

                strings.append(string)

            return ''.join(strings)


def _split_by_scriptencoding(bytes_seq):
    # type: (bytes) -> [(str, bytes)]
    max_end_index = len(bytes_seq)
    start_index = 0
    bytes_seq_and_loc_list = []

    while True:
        end_index = bytes_seq.find(SCRIPTENCODING_PREFIX, start_index + 1)

        if end_index < 0:
            end_index = max_end_index

        bytes_seq_and_loc_list.append((
            "{start_index}:{end_index}".format(start_index=start_index, end_index=end_index),
            bytes_seq[start_index:end_index]
        ))

        if end_index < max_end_index:
            start_index = end_index
            continue

        return bytes_seq_and_loc_list


class EncodingDetectionError(Exception):
    def __init__(self, debug_hint):
        # type: (Dict[str, Any]) -> None
        self.debug_hint = debug_hint


    def __str__(self):
        # type: () -> str
        return 'Cannot detect encoding (binary file?): {debug_hint}'.format(
            debug_hint=pformat(self.debug_hint)
        )



from vint.ast.node_type import NodeType
from vint.ast.traversing import traverse, register_traverser_extension
from vint.ast.parsing import Parser

STRING_EXPR_CONTENT = 'VINT:string_expression'
STRING_EXPR_CONTEXT = 'VINT:string_expression_context'



class MapAndFilterParser(object):
    """ A class to make string expression in map and filter function parseable. """
    def process(self, ast):
        def enter_handler(node):
            node_type = NodeType(node['type'])
            if node_type is not NodeType.CALL:
                return

            called_function_identifier = node['left']

            # Name node of the "map" or "filter" functions are always IDENTIFIER.
            if NodeType(called_function_identifier['type']) is not NodeType.IDENTIFIER:
                return

            is_map_or_function_call = called_function_identifier.get('value') in {
                'map': True,
                'filter': True,
            }

            if not is_map_or_function_call:
                return

            args = node['rlist']

            # Prevent crash. See https://github.com/Kuniwak/vint/issues/256.
            if len(args) < 2:
                return

            string_expr_node = args[1]

            # We can analyze only STRING nodes by static analyzing.
            if NodeType(string_expr_node['type']) is not NodeType.STRING:
                return

            string_expr_content_nodes = MapAndFilterParser._parse_string_expr_content_nodes(string_expr_node)
            node[STRING_EXPR_CONTENT] = string_expr_content_nodes

        traverse(ast, on_enter=enter_handler)

        return ast


    @classmethod
    def _parse_string_expr_content_nodes(cls, string_expr_node):
        parser = Parser()
        string_expr_content_nodes = parser.parse_string_expr(string_expr_node)

        def enter_handler(node):
            # NOTE: We need this flag only string nodes, because this flag is only for
            # ProhibitUnnecessaryDoubleQuote.
            if NodeType(node['type']) is NodeType.STRING:
                node[STRING_EXPR_CONTEXT] = {
                    'is_on_str_expr_context': True,
                }

        for string_expr_content_node in string_expr_content_nodes:
            traverse(string_expr_content_node, on_enter=enter_handler)

        return string_expr_content_nodes


def get_string_expr_content(node):
    return node.get(STRING_EXPR_CONTENT)


def get_string_context(node):
    return node.get(STRING_EXPR_CONTEXT, {
        'is_on_str_expr_context': False,
    })


@register_traverser_extension
def traverse_string_expr_content(node, on_enter=None, on_leave=None):
    string_expr_content_nodes = get_string_expr_content(node)
    if string_expr_content_nodes is None:
        return

    for child_node in string_expr_content_nodes:
        traverse(child_node, on_enter=on_enter, on_leave=on_leave)



# For dynamic policy module import
_policy_class_map = {}


def register_policy(PolicyClass):
    policy_name = PolicyClass.__name__
    _policy_class_map[policy_name] = PolicyClass
    return PolicyClass


def get_policy_class_map():
    return _policy_class_map



from vint.ast.node_type import NodeType
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy_registry import register_policy
from vint.ast.plugin.scope_plugin import (
    ExplicityOfScopeVisibility,
    ScopeVisibility,
)


@register_policy
class ProhibitImplicitScopeBuiltinVariable(AbstractPolicy):
    def __init__(self):
        super(ProhibitImplicitScopeBuiltinVariable, self).__init__()
        self.reference = ':help local-variable'
        self.level = Level.WARNING


    def listen_node_types(self):
        return [NodeType.IDENTIFIER]


    def is_valid(self, identifier, lint_context):
        """ Implicit scope builtin variables are prohibited.
        Because it will make unexpected variable name conflict between builtin
        and imlicit global/function local. For example:

            " This variable is not global variable but builtin variable.
            let count = 100
        """

        scope_plugin = lint_context['plugins']['scope']
        explicity = scope_plugin.get_explicity_of_scope_visibility(identifier)
        objective_scope_visibility = scope_plugin.get_objective_scope_visibility(identifier)

        is_valid = not (explicity is ExplicityOfScopeVisibility.IMPLICIT
                        and objective_scope_visibility is ScopeVisibility.BUILTIN)

        if not is_valid:
            self._make_description(identifier, scope_plugin)

        return is_valid


    def _make_description(self, identifier, scope_plugin):
        self.description = 'Make the scope explicit like `{good_example}` (or possibly be unexpected builtin?)'.format(
            good_example=scope_plugin.normalize_variable_name(identifier)
        )



import os
import os.path
from pathlib import Path
from vint.linting.file_filter import find_vim_script


def build_environment(cmdargs):
    return {
        'cmdargs': cmdargs,
        'home_path': _get_home_path(),
        'xdg_config_home': _get_xdg_config_home(),
        'cwd': _get_cwd(),
        'file_paths': _get_file_paths(cmdargs)
    }


def _get_cwd():
    return Path(os.getcwd())


def _get_home_path():
    return Path(os.path.expanduser('~'))


def _get_file_paths(cmdargs):
    if 'files' not in cmdargs:
        return []

    found_file_paths = find_vim_script(map(Path, cmdargs['files']))

    return set(found_file_paths)


def _get_xdg_config_home():
    return Path(os.environ.get(
        "XDG_CONFIG_HOME",
        str(_get_home_path().joinpath(".config"))
    ))



from enum import Enum, unique


@unique
class NodeType(Enum):
    TOPLEVEL = 1
    COMMENT = 2
    EXCMD = 3
    FUNCTION = 4
    ENDFUNCTION = 5
    DELFUNCTION = 6
    RETURN = 7
    EXCALL = 8
    LET = 9
    UNLET = 10
    LOCKVAR = 11
    UNLOCKVAR = 12
    IF = 13
    ELSEIF = 14
    ELSE = 15
    ENDIF = 16
    WHILE = 17
    ENDWHILE = 18
    FOR = 19
    ENDFOR = 20
    CONTINUE = 21
    BREAK = 22
    TRY = 23
    CATCH = 24
    FINALLY = 25
    ENDTRY = 26
    THROW = 27
    ECHO = 28
    ECHON = 29
    ECHOHL = 30
    ECHOMSG = 31
    ECHOERR = 32
    EXECUTE = 33
    TERNARY = 34
    OR = 35
    AND = 36
    EQUAL = 37
    EQUALCI = 38
    EQUALCS = 39
    NEQUAL = 40
    NEQUALCI = 41
    NEQUALCS = 42
    GREATER = 43
    GREATERCI = 44
    GREATERCS = 45
    GEQUAL = 46
    GEQUALCI = 47
    GEQUALCS = 48
    SMALLER = 49
    SMALLERCI = 50
    SMALLERCS = 51
    SEQUAL = 52
    SEQUALCI = 53
    SEQUALCS = 54
    MATCH = 55
    MATCHCI = 56
    MATCHCS = 57
    NOMATCH = 58
    NOMATCHCI = 59
    NOMATCHCS = 60
    IS = 61
    ISCI = 62
    ISCS = 63
    ISNOT = 64
    ISNOTCI = 65
    ISNOTCS = 66
    ADD = 67
    SUBTRACT = 68
    CONCAT = 69
    MULTIPLY = 70
    DIVIDE = 71
    REMAINDER = 72
    NOT = 73
    MINUS = 74
    PLUS = 75
    SUBSCRIPT = 76
    SLICE = 77
    CALL = 78
    DOT = 79
    NUMBER = 80
    STRING = 81
    LIST = 82
    DICT = 83
    NESTING = 84
    OPTION = 85
    IDENTIFIER = 86
    CURLYNAME = 87
    ENV = 88
    REG = 89
    CURLYNAMEPART = 90
    CURLYNAMEEXPR = 91
    LAMBDA = 92



import re
from vint.ast.node_type import NodeType
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy_registry import register_policy


@register_policy
class ProhibitSetNoCompatible(AbstractPolicy):
    def __init__(self):
        super(ProhibitSetNoCompatible, self).__init__()
        self.description = 'Do not use nocompatible which has unexpected effects'
        self.reference = ':help nocompatible'
        self.level = Level.WARNING


    def listen_node_types(self):
        return [NodeType.EXCMD]


    def is_valid(self, node, lint_context):
        """ Whether the specified node is valid.

        This policy prohibit following commands:
         - normal without !
         - substitute
        """

        command = node['str']
        is_nocompatible = re.match(r'set?\s+(?:nocp|invcp|nocompatible|invcompatible)', command)

        return not is_nocompatible



from vint.ast.traversing import traverse, SKIP_CHILDREN
from vint.ast.node_type import NodeType
from vint.ast.plugin.scope_plugin.scope_detector import (
    ScopeVisibility,
    detect_scope_visibility,
    is_analyzable_identifier,
    is_analyzable_declarative_identifier,
    is_builtin_variable,
    normalize_variable_name,
)
from vint.ast.plugin.scope_plugin.identifier_classifier import (
    IdentifierClassifier,
    is_function_identifier,
)


DeclarativeNodeTypes = {
    NodeType.FUNCTION: True,
    NodeType.LET: True,
    NodeType.FOR: True,
}


FunctionNameNodesDeclaringVariableSelf = {
    NodeType.DOT: True,
    NodeType.SUBSCRIPT: True,
}


SymbolTableVariableNames = {
    ScopeVisibility.GLOBAL_LIKE: ['g:', 'b:', 'w:', 't:', 'v:'],
    ScopeVisibility.SCRIPT_LOCAL: ['s:'],
    ScopeVisibility.FUNCTION_LOCAL: ['l:', 'a:'],
}



class ScopeLinker(object):
    """ A class for scope linkers.
    The class link identifiers in the given AST node and the scopes where the
    identifier will be declared or referenced.
    """

    class ScopeTreeBuilder(object):
        """ A class for event-driven builders to build a scope tree.
        The class interest to scope-level events rather than AST-level events.
        """

        def __init__(self):
            self.link_registry = ScopeLinker.ScopeLinkRegistry()

            global_scope = self._create_scope(ScopeVisibility.GLOBAL_LIKE)
            self._scope_stack = [global_scope]
            self._add_symbol_table_variables(global_scope)


        def enter_new_scope(self, scope_visibility):
            current_scope = self.get_current_scope()
            new_scope = self._create_scope(scope_visibility)
            self._add_symbol_table_variables(new_scope)

            # Build a lexical scope chain
            current_scope['child_scopes'].append(new_scope)
            self._scope_stack.append(new_scope)


        def leave_current_scope(self):
            self._scope_stack.pop()


        def get_global_scope(self):
            return self._scope_stack[0]


        def get_script_local_scope(self):
            return self._scope_stack[1]


        def get_current_scope(self):
            return self._scope_stack[-1]


        def handle_new_parameter_found(self, id_node):
            current_scope = self.get_current_scope()
            self._add_parameter(current_scope, id_node)


        def handle_new_range_parameters_found(self):
            # We can access "a:firstline" and "a:lastline" if the function is
            # declared with an attribute "range". See :func-range
            firstline_node = self._create_virtual_identifier('firstline')
            lastline_node = self._create_virtual_identifier('lastline')

            current_scope = self.get_current_scope()
            self._add_parameter(current_scope, firstline_node)
            self._add_parameter(current_scope, lastline_node)


        def handle_new_dict_parameter_found(self):
            # We can access "l:self" is declared with an attribute "dict".
            # See :help self
            current_scope = self.get_current_scope()
            self._add_self_variable(current_scope)


        def handle_new_parameters_list_and_length_found(self):
            # We can always access a:0 and a:000
            # See :help internal-variables
            param_length_node = self._create_virtual_identifier('0')
            param_list_node = self._create_virtual_identifier('000')

            current_scope = self.get_current_scope()
            self._add_parameter(current_scope, param_length_node)
            self._add_parameter(current_scope, param_list_node)


        def handle_new_index_parameters_found(self, params_number):
            current_scope = self.get_current_scope()

            # Max parameters number is 20. See :help E740
            for variadic_index in range(20 - params_number):
                # Variadic parameters named 1-based index.
                variadic_param = self._create_virtual_identifier(str(variadic_index + 1))
                self._add_parameter(current_scope, variadic_param)


        def handle_new_variable_found(self, node):
            current_scope = self.get_current_scope()
            scope_visibility_hint = detect_scope_visibility(
                node, current_scope)

            if scope_visibility_hint is None:
                # We cannot do anything
                return

            is_implicit = scope_visibility_hint['is_implicit']
            is_function = is_function_identifier(node)

            if is_builtin_variable(node):
                self._add_builtin_variable(node,
                                           is_function=is_function,
                                           is_implicit=is_implicit)
                return

            objective_scope = self._get_objective_scope(node)
            self._add_variable(objective_scope,
                               node,
                               is_function=is_function,
                               is_implicit=is_implicit)


        def _get_objective_scope(self, node):
            current_scope = self.get_current_scope()
            scope_visibility_hint = detect_scope_visibility(
                node, current_scope)
            scope_visibility = scope_visibility_hint['scope_visibility']

            if scope_visibility is ScopeVisibility.GLOBAL_LIKE:
                return self.get_global_scope()

            if scope_visibility is ScopeVisibility.SCRIPT_LOCAL:
                return self.get_script_local_scope()

            # It is FUNCTION_LOCAL scope
            return current_scope


        def handle_referencing_identifier_found(self, node):
            current_scope = self.get_current_scope()

            self.link_registry.link_identifier_to_context_scope(node, current_scope)


        def _add_parameter(self, objective_scope, node):
            variable_name = 'a:' + node['value']
            self._register_variable(objective_scope, variable_name, node)


        def _add_self_variable(self, objective_scope):
            variable_name = 'l:self'
            virtual_node = self._create_virtual_identifier(variable_name)
            self._register_variable(objective_scope, variable_name, virtual_node)


        def _add_variable(self, objective_scope, node, is_implicit=False, is_function=False):
            current_scope = self.get_current_scope()
            variable_name = normalize_variable_name(node, current_scope)

            self._register_variable(objective_scope,
                                    variable_name,
                                    node,
                                    is_function=is_function,
                                    is_implicit=is_implicit)


        def _add_builtin_variable(self, node, is_implicit=False, is_function=False):
            current_scope = self.get_current_scope()
            variable_name = normalize_variable_name(node, current_scope)

            self._register_variable(self.get_global_scope(),
                                    variable_name,
                                    node,
                                    is_implicit=is_implicit,
                                    is_function=is_function,
                                    is_builtin=True)
            return


        def _add_symbol_table_variables(self, objective_scope):
            # We can always access any symbol tables such as: "g:", "s:", "l:".
            # See :help internal-variables
            scope_visibility = objective_scope['scope_visibility']
            symbol_table_variable_names = SymbolTableVariableNames[scope_visibility]

            for symbol_table_variable_name in symbol_table_variable_names:
                virtual_node = self._create_virtual_identifier(symbol_table_variable_name)

                self._register_variable(objective_scope,
                                        symbol_table_variable_name,
                                        virtual_node)


        def _register_variable(self, objective_scope, variable_name, node,
                               is_implicit=False, is_builtin=False, is_function=False):
            variable = self._create_variable(is_implicit=is_implicit,
                                             is_builtin=is_builtin)

            objective_variable_list = objective_scope['functions' if is_function
                                                      else 'variables']

            same_name_variables = objective_variable_list.setdefault(variable_name, [])
            same_name_variables.append(variable)

            self.link_registry.link_variable_to_declarative_identifier(variable, node)

            current_scope = self.get_current_scope()
            self.link_registry.link_identifier_to_context_scope(node, current_scope)


        def _create_variable(self, is_implicit=False, is_builtin=False):
            return {
                'is_implicit': is_implicit,
                'is_builtin': is_builtin,
            }


        def _create_scope(self, scope_visibility):
            scope = {
                'scope_visibility': scope_visibility,
                'functions': {},
                'variables': {},
                'child_scopes': [],
            }

            return scope


        def _create_virtual_identifier(self, id_value):
            """ Returns a virtual identifier.
            Virtual identifier is a virtual node for implicitly declarated
            variables such as: a:0, a:000, a:firstline.
            """
            return {
                'type': NodeType.IDENTIFIER.value,
                'value': id_value,
                'is_virtual': True,
            }



    class ScopeLinkRegistry(object):
        """ A class for registry services for links between scopes and
        identifiers.
        """

        def __init__(self):
            self._vars_to_declarative_ids_map = {}
            self._ids_to_scopes_map = {}


        def link_variable_to_declarative_identifier(self, variable, declaring_id_node):
            variable_id = id(variable)
            self._vars_to_declarative_ids_map[variable_id] = declaring_id_node


        def link_identifier_to_context_scope(self, id_node, scope):
            node_id = id(id_node)
            self._ids_to_scopes_map[node_id] = scope


        def get_declarative_identifier_by_variable(self, variable):
            variable_id = id(variable)
            return self._vars_to_declarative_ids_map.get(variable_id)


        def get_context_scope_by_identifier(self, identifier):
            node_id = id(identifier)
            return self._ids_to_scopes_map.get(node_id)


    def __init__(self):
        self.scope_tree = None
        self.link_registry = None


    def process(self, ast):
        """ Build a scope tree and links between scopes and identifiers by the
        specified ast. You can access the built scope tree and the built links
        by .scope_tree and .link_registry.
        """
        id_classifier = IdentifierClassifier()
        attached_ast = id_classifier.attach_identifier_attributes(ast)

        self._scope_tree_builder = ScopeLinker.ScopeTreeBuilder()

        # We are already in script local scope.
        self._scope_tree_builder.enter_new_scope(ScopeVisibility.SCRIPT_LOCAL)

        traverse(attached_ast,
                 on_enter=self._enter_handler,
                 on_leave=self._leave_handler)

        self.scope_tree = self._scope_tree_builder.get_global_scope()
        self.link_registry = self._scope_tree_builder.link_registry


    def _find_variable_like_nodes(self, node):
        if not is_analyzable_identifier(node):
            return

        if is_analyzable_declarative_identifier(node):
            self._scope_tree_builder.handle_new_variable_found(node)
            return

        self._scope_tree_builder.handle_referencing_identifier_found(node)


    def _enter_handler(self, node):
        node_type = NodeType(node['type'])

        if node_type is NodeType.FUNCTION:
            return self._handle_function_node(node)

        self._find_variable_like_nodes(node)


    def _handle_function_node(self, func_node):
        # We should interrupt traversing, because a node of the function
        # name should be added to the parent scope before the current
        # scope switched to a new scope of the function.
        # We approach to it by the following 5 steps.
        #   1. Add the function to the current scope
        #   2. Create a new scope of the function
        #   3. The current scope point to the new scope
        #   4. Add parameters to the new scope
        #   5. Add variables in the function body to the new scope

        # 1. Add the function to the current scope
        func_name_node = func_node['left']
        traverse(func_name_node, on_enter=self._find_variable_like_nodes)

        # 2. Create a new scope of the function
        # 3. The current scope point to the new scope
        self._scope_tree_builder.enter_new_scope(ScopeVisibility.FUNCTION_LOCAL)

        has_variadic = False

        # 4. Add parameters to the new scope
        param_nodes = func_node['rlist']
        for param_node in param_nodes:
            if param_node['value'] == '...':
                has_variadic = True
            else:
                # the param_node type is always NodeType.IDENTIFIER
                self._scope_tree_builder.handle_new_parameter_found(param_node)

        # We can always access a:0, a:000
        self._scope_tree_builder.handle_new_parameters_list_and_length_found()

        # In a variadic function, we can access a:1 ... a:n
        # (n = 20 - explicit parameters length). See :help a:0
        if has_variadic:
            # -1 means ignore '...'
            self._scope_tree_builder.handle_new_index_parameters_found(len(param_nodes) - 1)

        # We can access "a:firstline" and "a:lastline" if the function is
        # declared with an attribute "range". See :func-range
        attr = func_node['attr']
        is_declared_with_range = attr['range'] is not 0
        if is_declared_with_range:
            self._scope_tree_builder.handle_new_range_parameters_found()

        # We can access "l:self" is declared with an attribute "dict" or
        # the function is a member of a dict. See :help self
        is_declared_with_dict = attr['dict'] is not 0 \
            or NodeType(func_name_node['type']) in FunctionNameNodesDeclaringVariableSelf
        if is_declared_with_dict:
            self._scope_tree_builder.handle_new_dict_parameter_found()

        # 5. Add variables in the function body to the new scope
        func_body_nodes = func_node['body']
        for func_body_node in func_body_nodes:
            traverse(func_body_node,
                     on_enter=self._enter_handler,
                     on_leave=self._leave_handler)

        # Skip child nodes traversing
        return SKIP_CHILDREN


    def _leave_handler(self, node):
        node_type = NodeType(node['type'])

        if node_type is NodeType.FUNCTION:
            self._scope_tree_builder.leave_current_scope()



from vint.linting.config.config_source import ConfigSource
from vint.linting.level import Level


class ConflictedOptionsError(Exception):
    def __init__(self, conflicted_option_names):
        self.conflicted_option_names = conflicted_option_names

    def __str__(self):
        option_names = ', '.join(self.conflicted_option_names)
        return 'These options are conflicted: {0}'.format(option_names)


class ConfigCmdargsSource(ConfigSource):
    def __init__(self, env):
        self._config_dict = self._build_config_dict(env)


    def get_config_dict(self):
        return self._config_dict


    def _build_config_dict(self, env):
        config_dict = {'cmdargs': {}}

        config_dict = self._normalize_color(env, config_dict)
        config_dict = self._normalize_json(env, config_dict)
        config_dict = self._normalize_stat(env, config_dict)
        config_dict = self._normalize_verbose(env, config_dict)
        config_dict = self._normalize_severity(env, config_dict)
        config_dict = self._normalize_max_violations(env, config_dict)
        config_dict = self._normalize_format(env, config_dict)
        config_dict = self._normalize_env(env, config_dict)

        return config_dict


    def _pass_config_by_key(self, key, env, config_dict):
        env_cmdargs = env['cmdargs']
        config_dict_cmdargs = config_dict['cmdargs']

        if key in env_cmdargs and env_cmdargs[key] is not None:
            config_dict_cmdargs[key] = env_cmdargs[key]

        return config_dict


    def _normalize_stat(self, env, config_dict):
        return self._pass_config_by_key('stat', env, config_dict)


    def _normalize_json(self, env, config_dict):
        return self._pass_config_by_key('json', env, config_dict)


    def _normalize_color(self, env, config_dict):
        env_cmdargs = env['cmdargs']

        should_colorize = 'color' in env_cmdargs and env_cmdargs['color'] is not None
        should_not_colorize = 'no_color' in env_cmdargs and env_cmdargs['no_color'] is not None

        if should_colorize and should_not_colorize:
            raise ConflictedOptionsError(['color', 'no_color'])

        if should_colorize:
            config_dict['cmdargs']['color'] = True

        if should_not_colorize:
            config_dict['cmdargs']['color'] = False

        return config_dict


    def _normalize_verbose(self, env, config_dict):
        return self._pass_config_by_key('verbose', env, config_dict)


    def _normalize_max_violations(self, env, config_dict):
        return self._pass_config_by_key('max-violations', env, config_dict)


    def _normalize_format(self, env, config_dict):
        return self._pass_config_by_key('format', env, config_dict)


    def _normalize_severity(self, env, config_dict):
        env_cmdargs = env['cmdargs']
        config_dict_cmdargs = config_dict['cmdargs']

        # Severity option priority is:
        #   1. error
        #   2. warning
        #   3. style problem
        if env_cmdargs.get('style_problem', False):
            config_dict_cmdargs['severity'] = Level.STYLE_PROBLEM

        if env_cmdargs.get('warning', False):
            config_dict_cmdargs['severity'] = Level.WARNING

        if env_cmdargs.get('error', False):
            config_dict_cmdargs['severity'] = Level.ERROR

        return config_dict


    def _normalize_env(self, env, config_dict):
        env_cmdargs = env['cmdargs']
        config_dict_cmdargs = config_dict['cmdargs']

        if 'enable_neovim' in env_cmdargs and env_cmdargs['enable_neovim'] is True:
            config_dict_cmdargs['env'] = {'neovim': True}

        return config_dict



import re
from vint.ast.node_type import NodeType
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy_registry import register_policy
from vint.ast.dictionary.autocmd_events import AutoCmdEvents


@register_policy
class ProhibitAutocmdWithNoGroup(AbstractPolicy):
    def __init__(self):
        super(ProhibitAutocmdWithNoGroup, self).__init__()
        self.description = 'autocmd should execute in an augroup or execute with a group'
        self.reference = ':help :autocmd'
        self.level = Level.WARNING

        self.is_inside_of_augroup = False


    def listen_node_types(self):
        return [NodeType.EXCMD]


    def is_valid(self, node, lint_context):
        """ Whether the specified node is valid.

        autocmd family should be called with any groups.
        """

        # noed.ea.cmd is empty when line jump command such as 1
        cmd_name = node['ea']['cmd'].get('name', None)

        is_autocmd = cmd_name == 'autocmd'
        if is_autocmd and not self.is_inside_of_augroup:
            matched = re.match(r'au(?:tocmd)?!?\s+(\S+)', node['str'])

            if not matched:
                # Looks like autocmd with a bang
                return True

            has_group = any(x and x.upper() not in AutoCmdEvents for x in matched.group(1).split(','))
            return has_group

        is_augroup = cmd_name == 'augroup'
        if is_augroup:
            matched = re.match(r'aug(?:roup)?\s+[eE][nN][dD]$', node['str'])
            is_augroup_end = bool(matched)
            self.is_inside_of_augroup = not is_augroup_end

        return True



from vint.ast.node_type import NodeType
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy.reference.googlevimscriptstyleguide import get_reference_source
from vint.linting.policy_registry import register_policy


ProhibitedCommands = {
    'substitute': True,
}

CommandsShouldBeWithBang = {
    'normal': True,
}


@register_policy
class ProhibitCommandRelyOnUser(AbstractPolicy):
    def __init__(self):
        super(ProhibitCommandRelyOnUser, self).__init__()
        self.description = 'Avoid commands that rely on user settings'
        self.reference = get_reference_source('FRAGILE')
        self.level = Level.WARNING


    def listen_node_types(self):
        return [NodeType.EXCMD]


    def is_valid(self, node, lint_context):
        """ Whether the specified node is valid.

        This policy prohibit following commands:
         - normal without !
         - substitute
        """

        ea = node['ea']
        command = ea['cmd'].get('name', None)

        # It seems line jump command
        if command is None:
            return True

        is_prohibited_command = command in ProhibitedCommands
        if is_prohibited_command:
            return False

        should_be_with_bang = command in CommandsShouldBeWithBang
        if not should_be_with_bang:
            return True

        is_bang = ea.get('forceit', 0) == 1
        return is_bang



from vint.ast.traversing import traverse
from vint.ast.plugin.scope_plugin.redir_assignment_parser import (
    RedirAssignmentParser,
    get_redir_content,
)
from vint.ast.plugin.scope_plugin.map_and_filter_parser import (
    MapAndFilterParser,
    get_string_expr_content,
)
from vint.ast.node_type import NodeType


IDENTIFIER_ATTRIBUTE = 'VINT:identifier_attribute'
IDENTIFIER_ATTRIBUTE_DECLARATION_FLAG = 'is_declarative'
IDENTIFIER_ATTRIBUTE_DYNAMIC_FLAG = 'is_dynamic'
IDENTIFIER_ATTRIBUTE_MEMBER_FLAG = 'is_member'
IDENTIFIER_ATTRIBUTE_FUNCTION_FLAG = 'is_function'
IDENTIFIER_ATTRIBUTE_AUTOLOAD_FLAG = 'is_autoload'
IDENTIFIER_ATTRIBUTE_PARAMETER_DECLARATION_FLAG = 'is_declarative_parameter'
IDENTIFIER_ATTRIBUTE_STRING_EXPRESSION_CONTEXT = 'is_on_str_expr_context'
REFERENCING_IDENTIFIERS = 'VINT:referencing_identifiers'
DECLARING_IDENTIFIERS = 'VINT:declaring_identifiers'


DeclarativeNodeTypes = {
    NodeType.LET: True,
    NodeType.FUNCTION: True,
    NodeType.FOR: True,
    NodeType.EXCMD: True,
}

IdentifierTerminateNodeTypes = {
    NodeType.IDENTIFIER: True,
    NodeType.ENV: True,
    NodeType.REG: True,
    NodeType.OPTION: True,
    NodeType.CURLYNAME: True,
}

AccessorLikeNodeTypes = {
    NodeType.DOT: True,
    NodeType.SUBSCRIPT: True,
    NodeType.SLICE: True,
}

# These node types is identifier-like when:
#   let object['name'] = 0
AnalyzableSubScriptChildNodeTypes = {
    NodeType.NUMBER: True,
    NodeType.STRING: True,
}


class IdentifierClassifier(object):
    """ A class for identifier classifiers.
    This class classify nodes by 5 flags:

    - is dynamic: True if the identifier name can be determined by static analysis.
    - is member: True if the identifier is a member of a subscription/dot/slice node.
    - is declaring: True if the identifier is used to declare.
    - is autoload: True if the identifier is declared with autoload.
    - is function: True if the identifier is a function. Vim distinguish
        between function identifiers and variable identifiers.
    - is declarative paramter: True if the identifier is a declarative
        parameter. For example, the identifier "param" in Func(param) is a
        declarative paramter.
    """

    class IdentifierCollector(object):
        """ A class for identifier node collectors.
        Only static and not member nodes will be collected and the nodes will
        be grouped by 2 types; declaring or referencing.
        """

        def collect_identifiers(self, ast):
            self.static_referencing_identifiers = []
            self.static_declaring_identifiers = []

            # TODO: Make more performance efficiency.
            traverse(ast, on_enter=self._enter_handler)

            return {
                'static_declaring_identifiers': self.static_declaring_identifiers,
                'static_referencing_identifiers': self.static_referencing_identifiers,
            }


        def _enter_handler(self, node):
            is_identifier_like_node = IDENTIFIER_ATTRIBUTE in node

            if not is_identifier_like_node:
                return

            id_attr = node[IDENTIFIER_ATTRIBUTE]
            if id_attr[IDENTIFIER_ATTRIBUTE_DYNAMIC_FLAG] or \
                    id_attr[IDENTIFIER_ATTRIBUTE_MEMBER_FLAG]:
                return

            if id_attr[IDENTIFIER_ATTRIBUTE_DECLARATION_FLAG]:
                self.static_declaring_identifiers.append(node)
            else:
                self.static_referencing_identifiers.append(node)


    def attach_identifier_attributes(self, ast):
        """ Attach 5 flags to the AST.

        - is dynamic: True if the identifier name can be determined by static analysis.
        - is member: True if the identifier is a member of a subscription/dot/slice node.
        - is declaring: True if the identifier is used to declare.
        - is autoload: True if the identifier is declared with autoload.
        - is function: True if the identifier is a function. Vim distinguish
            between function identifiers and variable identifiers.
        - is declarative paramter: True if the identifier is a declarative
            parameter. For example, the identifier "param" in Func(param) is a
            declarative paramter.
        - is on string expression context: True if the variable is on the
            string expression context. The string expression context is the
            string content on the 2nd argument of the map or filter function.
        """
        redir_assignment_parser = RedirAssignmentParser()
        ast_with_parsed_redir = redir_assignment_parser.process(ast)

        map_and_filter_parser = MapAndFilterParser()
        ast_with_parse_map_and_filter_and_redir = \
            map_and_filter_parser.process(ast_with_parsed_redir)

        traverse(ast_with_parse_map_and_filter_and_redir,
                 on_enter=self._enter_handler)
        return ast


    def _set_identifier_attribute(self, node, is_declarative=None, is_dynamic=None,
                                  is_member=None, is_function=None, is_autoload=None,
                                  is_declarative_parameter=None,
                                  is_on_str_expr_context=None):
        id_attr = node.setdefault(IDENTIFIER_ATTRIBUTE, {
            IDENTIFIER_ATTRIBUTE_DECLARATION_FLAG: False,
            IDENTIFIER_ATTRIBUTE_DYNAMIC_FLAG: False,
            IDENTIFIER_ATTRIBUTE_MEMBER_FLAG: False,
            IDENTIFIER_ATTRIBUTE_FUNCTION_FLAG: False,
            IDENTIFIER_ATTRIBUTE_AUTOLOAD_FLAG: False,
            IDENTIFIER_ATTRIBUTE_PARAMETER_DECLARATION_FLAG: False,
            IDENTIFIER_ATTRIBUTE_STRING_EXPRESSION_CONTEXT: False,
        })

        if is_declarative is not None:
            id_attr[IDENTIFIER_ATTRIBUTE_DECLARATION_FLAG] = is_declarative

        if is_dynamic is not None:
            id_attr[IDENTIFIER_ATTRIBUTE_DYNAMIC_FLAG] = is_dynamic

        if is_member is not None:
            id_attr[IDENTIFIER_ATTRIBUTE_MEMBER_FLAG] = is_member

        if is_function is not None:
            id_attr[IDENTIFIER_ATTRIBUTE_FUNCTION_FLAG] = is_function

        if is_autoload is not None:
            id_attr[IDENTIFIER_ATTRIBUTE_AUTOLOAD_FLAG] = is_autoload

        if is_declarative_parameter is not None:
            id_attr[IDENTIFIER_ATTRIBUTE_PARAMETER_DECLARATION_FLAG] = \
                is_declarative_parameter

        if is_on_str_expr_context is not None:
            id_attr[IDENTIFIER_ATTRIBUTE_STRING_EXPRESSION_CONTEXT] = \
                is_on_str_expr_context


    def _enter_handler(self, node):
        node_type = NodeType(node['type'])

        if node_type in IdentifierTerminateNodeTypes:
            # Attach identifier attributes to all IdentifierTerminateNodeTypes.
            self._set_identifier_attribute(node)

        if node_type in AccessorLikeNodeTypes:
            self._pre_mark_accessor_children(node)

        if node_type in DeclarativeNodeTypes:
            self._enter_declarative_node(node)

        if node_type is NodeType.CALL:
            self._enter_call_node(node)

        if node_type is NodeType.DELFUNCTION:
            self._enter_delfunction_node(node)
            return


    def _pre_mark_accessor_children(self, node):
        node_type = NodeType(node['type'])
        dict_node = node['left']

        if NodeType(dict_node['type']) in AccessorLikeNodeTypes:
            self._pre_mark_accessor_children(dict_node)

        if node_type is NodeType.SLICE:
            for member_node in node['rlist']:
                # In VimLParser spec, an empty array means null.
                #   list[1:] => {rlist: [node, []]}
                if type(member_node) is list:
                    continue

                if NodeType(member_node['type']) is NodeType.IDENTIFIER:
                    # Only the identifier should be flagged as a member that
                    # the variable is an accessor for a list or dictionary.
                    # For example, the variable that is "l:end" in list[0 : l:end]
                    # is not accessor for the symbol table of the variable "list",
                    # but it is a variable symbol table accessor.
                    continue

                self._pre_mark_member_node(member_node)
            return

        member_node = node['right']
        if node_type is NodeType.SUBSCRIPT:
            if NodeType(member_node['type']) is NodeType.IDENTIFIER:
                # Only the identifier should be flagged as a member that
                # the variable is an accessor for a list or dictionary.
                # For example, the variable that is "l:key" in dict[l:key]
                # is not accessor for the symbol table of the variable "dict",
                # but it is a variable symbol table accessor.
                return

        self._pre_mark_member_node(member_node)


    def _pre_mark_member_node(self, member_node):
        member_node_type = NodeType(member_node['type'])

        if member_node_type in IdentifierTerminateNodeTypes or \
                member_node_type in AnalyzableSubScriptChildNodeTypes:
            self._set_identifier_attribute(member_node, is_member=True)


    def _enter_identifier_like_node(self, node, is_declarative=None,
                                    is_function=None, is_declarative_parameter=None,
                                    is_on_str_expr_context=None):
        node_type = NodeType(node['type'])

        if node_type in AccessorLikeNodeTypes:
            id_like_node = node
            self._enter_accessor_node(id_like_node,
                                      is_declarative=is_declarative,
                                      is_function=is_function,
                                      is_declarative_parameter=is_declarative_parameter,
                                      is_on_str_expr_context=is_on_str_expr_context)
            return

        if node_type in IdentifierTerminateNodeTypes:
            id_like_node = node
            self._enter_identifier_node(id_like_node,
                                        is_declarative=is_declarative,
                                        is_function=is_function,
                                        is_declarative_parameter=is_declarative_parameter,
                                        is_on_str_expr_context=is_on_str_expr_context)
            return

        # Curlyname node can have a dynamic name. For example:
        #   let s:var = 'VAR'
        #   let my_{s:var} = 0
        if node_type is NodeType.CURLYNAME:
            id_like_node = node
            self._enter_curlyname_node(id_like_node,
                                       is_declarative=is_declarative,
                                       is_function=is_function,
                                       is_declarative_parameter=is_declarative_parameter,
                                       is_on_str_expr_context=is_on_str_expr_context)
            return


    def _enter_function_node(self, func_node):
        # Function node has declarative identifiers as the function name and
        # the paramerter names.

        # Function name is in the left.
        func_name_node = func_node['left']
        self._enter_identifier_like_node(func_name_node,
                                         is_declarative=True,
                                         is_function=True)

        # Function parameter names are in the r_list.
        func_param_nodes = func_node['rlist']
        for func_param_node in func_param_nodes:
            self._enter_identifier_like_node(func_param_node,
                                             is_declarative_parameter=True,
                                             is_declarative=True)


    def _enter_delfunction_node(self, delfunc_node):
        func_name_node = delfunc_node['left']
        self._enter_identifier_like_node(func_name_node,
                                         is_function=True)


    def _enter_curlyname_node(self, curlyname_node, is_declarative=None, is_function=None,
                              is_declarative_parameter=None, is_on_str_expr_context=None):
        self._set_identifier_attribute(curlyname_node,
                                       is_dynamic=True,
                                       is_declarative=is_declarative,
                                       is_function=is_function,
                                       is_declarative_parameter=is_declarative_parameter,
                                       is_on_str_expr_context=is_on_str_expr_context)


    def _enter_identifier_node(self, id_node, is_declarative=None, is_function=None,
                               is_declarative_parameter=None, is_on_str_expr_context=None):
        is_autoload = '#' in id_node['value']
        self._set_identifier_attribute(id_node,
                                       is_declarative=is_declarative,
                                       is_autoload=is_autoload,
                                       is_function=is_function,
                                       is_declarative_parameter=is_declarative_parameter,
                                       is_on_str_expr_context=is_on_str_expr_context)


    def _enter_accessor_node(self, accessor_node, is_declarative=None,
                             is_function=None, is_declarative_parameter=None,
                             is_on_str_expr_context=None):
        accessor_node_type = NodeType(accessor_node['type'])

        if accessor_node_type is NodeType.DOT:
            self._set_identifier_attribute(accessor_node['right'],
                                           is_declarative=is_declarative,
                                           is_dynamic=False,
                                           is_function=is_function,
                                           is_declarative_parameter=is_declarative_parameter,
                                           is_on_str_expr_context=is_on_str_expr_context)
            return

        if accessor_node_type is NodeType.SUBSCRIPT:
            subscript_right_type = NodeType(accessor_node['right']['type'])

            # We can do static analysis NodeType.SUBSCRIPT such as:
            #   let object['name'] = 0
            #
            # but we cannot do it in other cases such as:
            #   let object[var] = 0
            is_dynamic = subscript_right_type not in AnalyzableSubScriptChildNodeTypes

            if not is_dynamic:
                self._set_identifier_attribute(accessor_node['right'],
                                               is_declarative=is_declarative,
                                               is_dynamic=False,
                                               is_function=is_function,
                                               is_declarative_parameter=is_declarative_parameter,
                                               is_on_str_expr_context=is_on_str_expr_context)
            return

        if accessor_node_type is NodeType.SLICE:
            for elem_node in accessor_node['rlist']:
                if type(elem_node) is list:
                    # In VimLParser spec, an empty array means null.
                    #   list[1:] => {rlist: [node, []]}
                    continue

                elem_node_type = NodeType(elem_node['type'])

                # We can do static analysis NodeType.SLICE such as:
                #   let object[0:1] = 0
                #
                # but we cannot do it in other cases such as:
                #   let object[0:var] = 0
                is_dynamic = elem_node_type not in AnalyzableSubScriptChildNodeTypes

                # In the following case, 0 is a declarative but var is not declarative.
                # It is more like a reference.
                #   let object[0:var] = 0
                is_declarative = elem_node_type in AnalyzableSubScriptChildNodeTypes

                self._set_identifier_attribute(elem_node,
                                               is_declarative=is_declarative,
                                               is_dynamic=is_dynamic,
                                               is_function=is_function,
                                               is_declarative_parameter=is_declarative_parameter,
                                               is_on_str_expr_context=is_on_str_expr_context)
            return

        raise Exception()


    def _enter_let_node(self, let_node):
        # Only "=" operator can be used as declaration.
        if let_node['op'] != '=':
            return

        self._enter_destructuring_assignment_node(let_node)


    def _enter_for_node(self, for_node):
        self._enter_destructuring_assignment_node(for_node)


    def _enter_destructuring_assignment_node(self, node):
        # In VimLParser spec, an empty array means null.
        #
        # | Normal assignment    | Destructuring assignment    |
        # |:--------------------:|:---------------------------:|
        # | node['left'] == Node | node['left'] == []          |
        # | node['list'] == []   | node['list'] == [Node, ...] |

        left_node = node['left']
        is_destructuring_assignment = type(left_node) is list

        if is_destructuring_assignment:
            for elem_node in node['list']:
                self._enter_identifier_like_node(elem_node, is_declarative=True)
        else:
            self._enter_identifier_like_node(left_node, is_declarative=True)

        rest_node = node['rest']
        has_rest = type(rest_node) is not list
        if has_rest:
            self._enter_identifier_like_node(rest_node, is_declarative=True)


    def _enter_declarative_node(self, node):
        node_type = NodeType(node['type'])

        if node_type is NodeType.FUNCTION:
            self._enter_function_node(node)
            return
        if node_type is NodeType.LET:
            self._enter_let_node(node)
            return
        if node_type is NodeType.FOR:
            self._enter_for_node(node)
            return
        if node_type is NodeType.EXCMD:
            self._enter_excmd_node(node)
            return


    def _enter_call_node(self, call_node):
        left_node = call_node['left']
        self._enter_identifier_like_node(left_node, is_function=True)

        # Care the 2nd argument of the map or filter function.
        self._enter_str_expr_content_node(call_node)


    def _enter_str_expr_content_node(self, call_node):
        string_expr_content_nodes = get_string_expr_content(call_node)
        if not string_expr_content_nodes:
            return

        def enter_handler(node):
            self._enter_identifier_like_node(node, is_on_str_expr_context=True)

        for string_expr_content_node in string_expr_content_nodes:
            traverse(string_expr_content_node, on_enter=enter_handler)


    def _enter_excmd_node(self, cmd_node):
        # Care an assignment by using command ":redir"
        redir_content_node = get_redir_content(cmd_node)

        if not redir_content_node:
            return

        self._enter_identifier_like_node(redir_content_node, is_declarative=True)



def is_identifier_like_node(node):
    return IDENTIFIER_ATTRIBUTE in node


def is_function_identifier(node):
    if not is_identifier_like_node(node):
        return False

    return node[IDENTIFIER_ATTRIBUTE][IDENTIFIER_ATTRIBUTE_FUNCTION_FLAG]


def is_dynamic_identifier(node):
    if not is_identifier_like_node(node):
        return False

    return node[IDENTIFIER_ATTRIBUTE][IDENTIFIER_ATTRIBUTE_DYNAMIC_FLAG]


def is_declarative_identifier(node):
    if not is_identifier_like_node(node):
        return False

    return node[IDENTIFIER_ATTRIBUTE][IDENTIFIER_ATTRIBUTE_DECLARATION_FLAG]


def is_member_identifier(node):
    if not is_identifier_like_node(node):
        return False

    return node[IDENTIFIER_ATTRIBUTE][IDENTIFIER_ATTRIBUTE_MEMBER_FLAG]


def is_autoload_identifier(node):
    if not is_identifier_like_node(node):
        return False

    return node[IDENTIFIER_ATTRIBUTE][IDENTIFIER_ATTRIBUTE_AUTOLOAD_FLAG]


def is_declarative_parameter(node):
    if not is_identifier_like_node(node):
        return False

    return node[IDENTIFIER_ATTRIBUTE][IDENTIFIER_ATTRIBUTE_PARAMETER_DECLARATION_FLAG]


def is_on_string_expression_context(node):
    if not is_identifier_like_node(node):
        return False

    return node[IDENTIFIER_ATTRIBUTE][IDENTIFIER_ATTRIBUTE_STRING_EXPRESSION_CONTEXT]



import re
from vint.ast.node_type import NodeType
from vint.ast.plugin.scope_plugin.map_and_filter_parser import get_string_context
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy.reference.googlevimscriptstyleguide import get_reference_source
from vint.linting.policy_registry import register_policy



# see `:help expr-string`
_special_char_matcher = re.compile(
    r'(\'|'  # allow single quote
    r'\\('  # prefix back slash
    r'(?P<octal>[0-7]{1,3})'
    r'|(?P<hexadecimal>[xX][0-9a-fA-F]{1,2})'
    r'|(?P<numeric_character_reference>[uU][0-9a-fA-F]{4})'
    r'|(?P<backspace>b)'
    r'|(?P<escape>e)'
    r'|(?P<form_feed>f)'
    r'|(?P<new_line>n)'
    r'|(?P<carriage_return>r)'
    r'|(?P<tab>t)'
    r'|(?P<backslash>\\)'
    r'|(?P<double_quote>")'
    r'|(?P<special_key><[^>]+>)'
    r'))')


@register_policy
class ProhibitUnnecessaryDoubleQuote(AbstractPolicy):
    def __init__(self):
        super(ProhibitUnnecessaryDoubleQuote, self).__init__()
        self.description = 'Prefer single quoted strings'
        self.reference = get_reference_source('STRINGS')
        self.level = Level.WARNING


    def listen_node_types(self):
        return [NodeType.STRING]


    def is_valid(self, node, lint_context):
        """ Whether the specified node is valid.

        In this policy, valid node is only 2 cases;
        - single quoted
        - double quoted, but including a special char

        See `:help expr-string`. """

        value = node['value']

        is_double_quoted = value[0] == '"'
        if not is_double_quoted:
            return True

        has_escaped_char = _special_char_matcher.search(value) is not None

        if has_escaped_char:
            return True

        string_expr_context = get_string_context(node)
        return string_expr_context['is_on_str_expr_context']



def get_config_value(config_dict, keys, default=None):
    inner_dict = config_dict

    for key in keys[:-1]:
        if key not in inner_dict:
            return default

        inner_dict = inner_dict[key]

    return inner_dict.get(keys[-1], default)



from __future__ import absolute_import

try:
    # Try to import system-wide vimlparser module first
    import vimlparser
except ImportError:
    # Fallback to the bundled one if not found
    from . import vimlparser



class AbstractASTPlugin(object):
    """ An abstract class for AST plugins. AST plugins can add attributes to
    the AST. But for maintainability reason, overwriting or deleting any
    attribute is prohibited.
    """

    def process(self, ast):
        return ast



from vint.ast.traversing import traverse, register_traverser_extension

from vint.ast.parsing import Parser
from vint.ast.node_type import NodeType

REDIR_CONTENT = 'VINT:redir_content'



class RedirAssignmentParser(object):
    """ A class to make redir assignment parseable. """


    def process(self, ast):
        def enter_handler(node):
            node_type = NodeType(node['type'])
            if node_type is not NodeType.EXCMD:
                return

            is_redir_command = node['ea']['cmd'].get('name') == 'redir'
            if not is_redir_command:
                return

            redir_cmd_str = node['str']
            is_redir_assignment = '=>' in redir_cmd_str
            if not is_redir_assignment:
                return

            parser = Parser()
            redir_content_node = parser.parse_redir(node)
            node[REDIR_CONTENT] = redir_content_node

        traverse(ast, on_enter=enter_handler)

        return ast


def get_redir_content(node):
    return node.get(REDIR_CONTENT)



@register_traverser_extension
def traverse_redir_content(node, on_enter=None, on_leave=None):
    if REDIR_CONTENT not in node:
        return

    traverse(node[REDIR_CONTENT], on_enter=on_enter, on_leave=on_leave)



import chardet
from typing import Optional, Dict, Any


SCRIPTENCODING_PREFIX = bytearray('scriptencoding', encoding='ascii')
COMMENT_START_TOKEN = bytearray('"', encoding='ascii')
LF = bytearray("\n", encoding='ascii')


class DecodingStrategy(object):
    def decode(self, bytes_seq, debug_hint):
        # type: (bytes, Dict[str, str]) -> Optional[str]
        raise NotImplementedError


class DecodingStrategyByChardet(DecodingStrategy):
    def decode(self, bytes_seq, debug_hint):
        # type: (bytes, Dict[str, Any]) -> Optional[str]
        encoding_hint = chardet.detect(bytearray(bytes_seq))
        encoding = encoding_hint['encoding']

        debug_hint['chardet_encoding'] = encoding_hint['encoding']
        debug_hint['chardet_confidence'] = encoding_hint['confidence']

        try:
            return bytes_seq.decode(encoding)

        except Exception as e:
            debug_hint['chardet_error'] = str(e)
            return None


class ComposedDecodingStrategy(DecodingStrategy):
    def __init__(self, strategies):
        # type: ([DecodingStrategy]) -> None
        self.strategies = strategies


    def decode(self, bytes_seq, debug_hint):
        # type: (bytes, Dict[str, Any]) -> Optional[str]

        debug_hint['composed_strategies'] = [type(strategy).__name__ for strategy in self.strategies]

        for strategy in self.strategies:
            string_candidate = strategy.decode(bytes_seq, debug_hint)

            if string_candidate is None:
                continue

            debug_hint['selected_strategy'] = type(strategy).__name__

            return string_candidate


class DecodingStrategyForEmpty(DecodingStrategy):
    def decode(self, bytes_seq, debug_hint):
        # type: (bytes, Dict[str, Any]) -> Optional[str]
        if len(bytes_seq) <= 0:
            debug_hint['empty'] = 'true'
            return ''

        debug_hint['empty'] = 'false'
        return None


class DecodingStrategyByScriptencoding(DecodingStrategy):
    def decode(self, bytes_seq, debug_hint):
        # type: (bytes, Dict[str, Any]) -> Optional[str]
        encoding_part = DecodingStrategyByScriptencoding.parse_script_encoding(bytes_seq, debug_hint)

        if encoding_part is None:
            debug_hint['scriptencoding'] = 'None'
            return None

        try:
            debug_hint['scriptencoding'] = encoding_part
            return bytes_seq.decode(encoding=encoding_part.decode(encoding='ascii'))

        except LookupError as e:
            debug_hint['scriptencoding_error'] = str(e)
            return None


    @classmethod
    def parse_script_encoding(cls, bytes_seq, debug_hint):
        # type: (bytes, Dict[str, Any]) -> Optional[bytes]
        try:
            start_index = bytes_seq.index(SCRIPTENCODING_PREFIX)
            encoding_part_start_index = start_index + len(SCRIPTENCODING_PREFIX)

            try:
                encoding_part_end_index_candidate_by_line_break = bytes_seq.index(LF, encoding_part_start_index)

                try:
                    encoding_part_end_index_candidate_by_comment = bytes_seq.index(
                        COMMENT_START_TOKEN, encoding_part_start_index)

                    # Case for :scriptencoding foo "foo\n
                    encoding_part_end_index = min(
                        encoding_part_end_index_candidate_by_line_break,
                        encoding_part_end_index_candidate_by_comment
                    )

                except ValueError:
                    # Case for :scriptencoding foo\n
                    encoding_part_end_index = encoding_part_end_index_candidate_by_line_break

            except ValueError:
                try:
                    # Case for :scriptencoding foo "foo<EOF>
                    encoding_part_end_index_candidate_by_comment = bytes_seq.index(
                        COMMENT_START_TOKEN, encoding_part_start_index)
                    encoding_part_end_index = encoding_part_end_index_candidate_by_comment

                except ValueError:
                    # Case for :scriptencoding foo<EOF>
                    encoding_part_end_index = len(bytes_seq) - 1

            encoding_part_candidate = bytes_seq[encoding_part_start_index:encoding_part_end_index]
            return encoding_part_candidate.strip()

        except ValueError:
            debug_hint['scriptencoding_error'] = '`scriptencoding` is not found'
            return None


class DecodingStrategyForUTF8(DecodingStrategy):
    def decode(self, bytes_seq, debug_hint):
        # type: (bytes, Dict[str, Any]) -> Optional[str]
        try:
            string = bytes_seq.decode('utf8')

            debug_hint['utf-8'] = 'success'
            return string

        except Exception as e:
            debug_hint['utf-8'] = 'failed: {}'.format(str(e))

            return None


default_decoding_strategy = ComposedDecodingStrategy([
    DecodingStrategyForEmpty(),
    DecodingStrategyByScriptencoding(),
    DecodingStrategyForUTF8(),
    DecodingStrategyByChardet(),
])
